From daf1ea20c37b37182a0eed1f7c6077889c3d4ce2 Mon Sep 17 00:00:00 2001
From: tanghy <tanghy.fnst@fujitsu.com>
Date: Thu, 20 Jan 2022 16:08:52 +0900
Subject: [PATCH v12] Support tab completion with a query result for upper


diff --git a/src/bin/psql/t/010_tab_completion.pl b/src/bin/psql/t/010_tab_completion.pl
index d3d1bd650e..eb334b59f6 100644
--- a/src/bin/psql/t/010_tab_completion.pl
+++ b/src/bin/psql/t/010_tab_completion.pl
@@ -33,6 +33,11 @@ if ($@)
 	plan skip_all => 'IO::Pty is needed to run this test';
 }
 
+# Support tab completion for upper character input in single byte encoding.
+# Use LATIN1 to test that.
+$ENV{LC_ALL}           = 'C';
+$ENV{PGCLIENTENCODING} = 'LATIN1';
+
 # start a new server
 my $node = PostgreSQL::Test::Cluster->new('main');
 $node->init;
@@ -41,9 +46,13 @@ $node->start;
 # set up a few database objects
 $node->safe_psql('postgres',
 	    "CREATE TABLE tab1 (f1 int, f2 text);\n"
+	  . "CREATE TABLE onetab1 (f1 int);\n"
 	  . "CREATE TABLE mytab123 (f1 int, f2 text);\n"
 	  . "CREATE TABLE mytab246 (f1 int, f2 text);\n"
-	  . "CREATE TYPE enum1 AS ENUM ('foo', 'bar', 'baz');\n");
+	  . "CREATE TABLE \"myTAB123\" (\"aF1\" int, f2 text);\n"
+	  . "CREATE TYPE enum1 AS ENUM ('foo', 'bar', 'baz');\n"
+	  . "CREATE TYPE myenum1 as enum ('bLACK');\n");
+
 
 # Developers would not appreciate this test adding a bunch of junk to
 # their ~/.psql_history, so be sure to redirect history into a temp file.
@@ -150,6 +159,66 @@ check_completion("SEL\t", qr/SELECT /, "complete SEL<tab> to SELECT");
 
 clear_query();
 
+# check set query command(upper case) completion for upper character inputs
+check_completion("set BYT\t", qq/set BYT\b\b\bbytea_output /, "complete set BYT<tab> to set bytea_output");
+
+clear_query();
+
+# check set query command(lower case) completion for upper character inputs
+check_completion("set bYT\t", qq/set bYT\b\bytea_output /, "complete set bYT<tab> to set bytea_output");
+
+clear_query();
+
+# check query command(upper case) completion for empty input
+check_completion("update onetab1 \t", qr/update onetab1 SET /, "complete SQL key words for onetab1 with empty input");
+
+clear_query();
+
+# check query command(lower case) completion for empty input
+check_completion("update onetab1 SET \t", qr/update onetab1 SET f1 /, "complete column name for onetab1 with empty input");
+
+clear_query();
+
+# check query command completion for upper character relation name
+check_completion("update TAB1 SET \t", qr/update TAB1 SET \af/, "complete column name for TAB1");
+
+clear_query();
+
+# check quoted identifiers in table
+check_completion("update \"my\t", qr/update \"myTAB123\" /, "complete quoted string1");
+
+clear_query();
+
+# check quoted identifiers in column
+check_completion("update \"myTAB123\" SET \"aF\t", qr/update \"myTAB123\" SET \"aF1\" /, "complete quoted string2");
+
+clear_query();
+
+# check schema query(lower case) which is case-insensitive
+check_completion("select oid from pg_Cla\t", qq/select oid from pg_Cla\b\b\bclass /, "complete schema query with lower case string");
+
+clear_query();
+
+# check schema query(upper case) which is case-insensitive
+check_completion("select oid from Pg_cla\t", qq/select oid from Pg_cla\b\b\b\b\b\bpg_class /, "complete schema query with uppper case string");
+
+clear_query();
+
+# check schema.table query which is case-insensitive
+check_completion("alter table PUBLIC.tab\t", qq/alter table PUBLIC.tab\b\b\b\b\b\b\b\b\b\bpublic.tab1 /, "complete schema.table without quoted identifiers");
+
+clear_query();
+
+# check schema.table query which is case-sensitive
+check_completion("alter table PUBLIC.\"my\t", qq/alter table PUBLIC.\"my\b\b\b\b\b\b\b\b\b\bpublic.\"myTAB123\" /, "complete schema.table with quoted identifiers");
+
+clear_query();
+
+# check enum values which are case-insensitive
+check_completion("ALTER TYPE myenum1 RENAME VALUE '\t\t", qr|'bLACK'|, "complete enum values");
+
+clear_line();
+
 # check case variation is honored
 check_completion("sel\t", qr/select /, "complete sel<tab> to select");
 
diff --git a/src/bin/psql/tab-complete.c b/src/bin/psql/tab-complete.c
index 6bd33a06cb..5b56058973 100644
--- a/src/bin/psql/tab-complete.c
+++ b/src/bin/psql/tab-complete.c
@@ -47,6 +47,7 @@
 #include "catalog/pg_class_d.h"
 #include "common.h"
 #include "libpq-fe.h"
+#include "mb/pg_wchar.h"
 #include "pqexpbuffer.h"
 #include "settings.h"
 #include "stringutils.h"
@@ -1062,7 +1063,7 @@ static const VersionedQuery Query_for_list_of_subscriptions[] = {
 };
 
 /*
- * This is a list of all "things" in Pgsql, which can show up after CREATE or
+ * This is a list of all "things" in pgsql, which can show up after CREATE or
  * DROP; and there is also a query to get a list of them.
  */
 
@@ -1212,6 +1213,7 @@ static char *complete_from_files(const char *text, int state);
 
 static char *pg_strdup_keyword_case(const char *s, const char *ref);
 static char *escape_string(const char *text);
+static char *pg_string_tolower_if_single_byte(const char *text);
 static PGresult *exec_query(const char *query);
 
 static char **get_previous_words(int point, char **buffer, int *nwords);
@@ -4651,6 +4653,7 @@ _complete_from_query(const char *simple_query,
 		char	   *e_text;
 		char	   *e_info_charp;
 		char	   *e_info_charp2;
+		char	   *le_str;
 		const char *pstr = text;
 		int			char_length = 0;
 
@@ -4674,13 +4677,27 @@ _complete_from_query(const char *simple_query,
 		/* Set up suitably-escaped copies of textual inputs */
 		e_text = escape_string(text);
 
+		/* Change the textual inputs to lower case if they're in single byte*/
+		le_str = pg_string_tolower_if_single_byte(e_text);
+		free(e_text);
+		e_text = le_str;
 		if (completion_info_charp)
+		{
 			e_info_charp = escape_string(completion_info_charp);
+			le_str = pg_string_tolower_if_single_byte(e_info_charp);
+			free(e_info_charp);
+			e_info_charp = le_str;
+		}
 		else
 			e_info_charp = NULL;
 
 		if (completion_info_charp2)
+		{
 			e_info_charp2 = escape_string(completion_info_charp2);
+			le_str = pg_string_tolower_if_single_byte(e_info_charp2);
+			free(e_info_charp2);
+			e_info_charp2 = le_str;
+		}
 		else
 			e_info_charp2 = NULL;
 
@@ -4715,7 +4732,7 @@ _complete_from_query(const char *simple_query,
 			 */
 			if (strcmp(schema_query->catname,
 					   "pg_catalog.pg_class c") == 0 &&
-				strncmp(text, "pg_", 3) != 0)
+				pg_strncasecmp(text, "pg_", 3) != 0)
 			{
 				appendPQExpBufferStr(&query_buffer,
 									 " AND c.relnamespace <> (SELECT oid FROM"
@@ -5144,6 +5161,39 @@ escape_string(const char *text)
 }
 
 
+/*
+ * pg_string_tolower_if_single_byte - Fold a string to lower case if the string is
+ * not quoted and in single byte encoding.
+ *
+ * The returned value has to be freed.
+ */
+static char *
+pg_string_tolower_if_single_byte(const char *text)
+{
+	char	*ret,
+			*p;
+	bool	enc_is_single_byte;
+	bool	notquoted = true;
+	int		count = 0;
+
+	ret = pg_strdup(text);
+	enc_is_single_byte = pg_wchar_table[pset.encoding].maxmblen == 1;
+
+	for (p = ret; *p; p++)
+	{
+		if (p[0] == '"')
+		{
+			count++;
+			notquoted = false;
+		}
+		if (enc_is_single_byte && notquoted && (count % 2 == 0))
+			*p = pg_tolower((unsigned char) *p);
+	}
+
+	return ret;
+}
+
+
 /*
  * Execute a query and report any errors. This should be the preferred way of
  * talking to the database in this file.
-- 
2.33.0.windows.2

